#include <ofxCore.h>
#include <ofxProperty.h>
#include <ofxParam.h>
//#include <ofxMultiThread.h>
//#include <ofxInteract.h>
//#include <ofxKeySyms.h>
//#include <ofxMemory.h>
//#include <ofxMessage.h>
#include <ofxImageEffect.h>
#include <ofxPixels.h>

#define NOMINMAX
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>

#include <algorithm>
#include <string>

#include <cstring>
#include <cassert>
#include <cstdint>

#ifdef _DEBUG
#define OFX_ASSERT_OK(expr)   do { OfxStatus status__ = (expr); assert(status__ == kOfxStatOK); } while(false)
#else
#define OFX_ASSERT_OK(expr)   do { (expr); } while(false)
#endif

static const char* const EFFECT_GROUP_NAME = "OFX Example";
static const char* const EFFECT_NAME = "Test1";

static OfxHost* g_Host;
static OfxImageEffectSuiteV1* g_EffectSuite;
static OfxPropertySuiteV1* g_PropertySuite;
static OfxParameterSuiteV1* g_ParameterSuite;

class EffectInstance
{
public:
    OfxParamHandle SizeParam;
    OfxParamHandle SpeedParam;
    OfxParamHandle IntensityParam;

    EffectInstance();
    ~EffectInstance();

private:
};

EffectInstance::EffectInstance() :
    SizeParam(nullptr),
    SpeedParam(nullptr),
    IntensityParam(nullptr)
{
}

EffectInstance::~EffectInstance()
{
}

static OfxStatus OnLoad()
{
    if(!g_Host)
        return kOfxStatErrMissingHostFeature;
    
    g_EffectSuite = (OfxImageEffectSuiteV1*)g_Host->fetchSuite(g_Host->host, kOfxImageEffectSuite, 1);
    g_PropertySuite = (OfxPropertySuiteV1*)g_Host->fetchSuite(g_Host->host, kOfxPropertySuite, 1);
    g_ParameterSuite = (OfxParameterSuiteV1*)g_Host->fetchSuite(g_Host->host, kOfxParameterSuite, 1);
    
    if(!g_EffectSuite || !g_PropertySuite || !g_ParameterSuite)
        return kOfxStatErrMissingHostFeature;

    return kOfxStatOK;
}

static OfxStatus OnUnload()
{
    return kOfxStatOK;
}

static OfxStatus OnDescribe(OfxImageEffectHandle effectDescriptor)
{
    assert(g_EffectSuite && g_PropertySuite);

    OfxPropertySetHandle effectDescriptorProperties = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->getPropertySet(effectDescriptor, &effectDescriptorProperties) );

    OFX_ASSERT_OK( g_PropertySuite->propSetString(effectDescriptorProperties, kOfxPropLabel, 0, EFFECT_NAME) );
    OFX_ASSERT_OK( g_PropertySuite->propSetString(effectDescriptorProperties, kOfxImageEffectPluginPropGrouping, 0, EFFECT_GROUP_NAME) );
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPluginPropSingleInstance, 0, 0) ); // Default
    OFX_ASSERT_OK( g_PropertySuite->propSetString(effectDescriptorProperties, kOfxImageEffectPropSupportedContexts, 0, kOfxImageEffectContextFilter) );
    OFX_ASSERT_OK( g_PropertySuite->propSetString(effectDescriptorProperties, kOfxImageEffectPluginRenderThreadSafety, 0, kOfxImageEffectRenderFullySafe) );
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPluginPropHostFrameThreading, 0, 1) ); // Default
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPropSupportsMultiResolution, 0, 1) ); // Default
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPropTemporalClipAccess, 0, 0) ); // Default. 1 indicates the plugin needs temporal access to images.
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPluginPropFieldRenderTwiceAlways, 0, 1) ); // Default.
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPropSupportsMultipleClipDepths, 0, 0) ); // Default
    //OFX_ASSERT_OK( g_PropertySuite->propSetInt(effectDescriptorProperties, kOfxImageEffectPropSupportsMultipleClipPARs, 0, 0) ); // Default
    
    const char* supportedPixelDepths[] = {
        kOfxBitDepthByte,
        //kOfxBitDepthShort,
        //kOfxBitDepthFloat,
    };
    OFX_ASSERT_OK( g_PropertySuite->propSetStringN(effectDescriptorProperties, kOfxImageEffectPropSupportedPixelDepths, _countof(supportedPixelDepths), supportedPixelDepths) );
  
    return kOfxStatOK;
}

static OfxStatus OnDescribeInContext(OfxImageEffectHandle effectDescriptor, OfxPropertySetHandle inArgs)
{
    assert(g_EffectSuite && g_PropertySuite);

    OfxPropertySetHandle props = nullptr;

    // define the single output clip in both contexts
    OFX_ASSERT_OK( g_EffectSuite->clipDefine(effectDescriptor, "Output", &props) );

    // set the component types we can handle on our output
    OFX_ASSERT_OK( g_PropertySuite->propSetString(props, kOfxImageEffectPropSupportedComponents, 0, kOfxImageComponentRGBA) );

    // define the single source clip in both contexts
    OFX_ASSERT_OK( g_EffectSuite->clipDefine(effectDescriptor, "Source", &props) );

    // set the component types we can handle on our main input
    OFX_ASSERT_OK( g_PropertySuite->propSetString(props, kOfxImageEffectPropSupportedComponents, 0, kOfxImageComponentRGBA) );

    return kOfxStatOK;
}

static inline OfxRGBAColourB* pixelAddress(OfxRGBAColourB *img, OfxRectI rect, int x, int y, int bytesPerLine)
{  
    if(x < rect.x1 || x >= rect.x2 || y < rect.y1 || y > rect.y2)
        return 0;
    OfxRGBAColourB *pix = (OfxRGBAColourB*)(((char*) img) + (y - rect.y1) * bytesPerLine);
    pix += x - rect.x1;  
    return pix;
}

static OfxStatus OnRender(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    assert(g_EffectSuite && g_PropertySuite);

    OfxPropertySetHandle effectInstanceProperties = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->getPropertySet(effectInstance, &effectInstanceProperties) );
    EffectInstance* instanceObj = nullptr;
    OFX_ASSERT_OK( g_PropertySuite->propGetPointer(effectInstanceProperties, kOfxPropInstanceData, 0, (void**)&instanceObj) );
    assert(instanceObj->m_Test == 123);

    // get the render window and the time from the inArgs
    OfxTime time = 0.0;
    OfxRectI renderWindow = { 0 };
  
    OFX_ASSERT_OK( g_PropertySuite->propGetDouble(inArgs, kOfxPropTime, 0, &time) );
    OFX_ASSERT_OK( g_PropertySuite->propGetIntN(inArgs, kOfxImageEffectPropRenderWindow, 4, &renderWindow.x1) );

    double size = 0.f, speed = 0.f, intensity = 0.f;
    OFX_ASSERT_OK( g_ParameterSuite->paramGetValueAtTime(instanceObj->SizeParam, time, &size) );
    OFX_ASSERT_OK( g_ParameterSuite->paramGetValueAtTime(instanceObj->SpeedParam, time, &speed) );
    OFX_ASSERT_OK( g_ParameterSuite->paramGetValueAtTime(instanceObj->IntensityParam, time, &intensity) );

    // fetch output clip
    OfxImageClipHandle outputClip = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->clipGetHandle(effectInstance, "Output", &outputClip, 0) );

    // fetch image to render into from that clip
    OfxPropertySetHandle outputImg = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->clipGetImage(outputClip, time, NULL, &outputImg) );

    // fetch output image info from that handle
    int dstRowBytes = 0, dstBitDepth = 0;
    OfxRectI dstRect = { 0 };
    void *dstPtr = nullptr;
    OFX_ASSERT_OK( g_PropertySuite->propGetInt(outputImg, kOfxImagePropRowBytes, 0, &dstRowBytes) );
    OFX_ASSERT_OK( g_PropertySuite->propGetIntN(outputImg, kOfxImagePropBounds, 4, &dstRect.x1) );
    OFX_ASSERT_OK( g_PropertySuite->propGetPointer(outputImg, kOfxImagePropData, 0, &dstPtr) );
  
    // fetch main input clip
    OfxImageClipHandle sourceClip = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->clipGetHandle(effectInstance, "Source", &sourceClip, 0) );

    // fetch image at render time from that clip
    OfxPropertySetHandle sourceImg = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->clipGetImage(sourceClip, time, NULL, &sourceImg) );

    // fetch image info out of that handle
    int srcRowBytes = 0, srcBitDepth = 0;
    OfxRectI srcRect = { 0 };
    void *srcPtr = nullptr;
    OFX_ASSERT_OK( g_PropertySuite->propGetInt(sourceImg, kOfxImagePropRowBytes, 0, &srcRowBytes) );
    OFX_ASSERT_OK( g_PropertySuite->propGetIntN(sourceImg, kOfxImagePropBounds, 4, &srcRect.x1) );
    OFX_ASSERT_OK( g_PropertySuite->propGetPointer(sourceImg, kOfxImagePropData, 0, &srcPtr) );

    // cast data pointers to 8 bit RGBA
    OfxRGBAColourB *src = (OfxRGBAColourB*)srcPtr;
    OfxRGBAColourB *dst = (OfxRGBAColourB*)dstPtr;

    for(int y = renderWindow.y1; y < renderWindow.y2; y++)
    {
        if(g_EffectSuite->abort(effectInstance)) break;
        OfxRGBAColourB *dstPix = pixelAddress(dst, dstRect, renderWindow.x1, y, dstRowBytes);
        OfxRGBAColourB *srcPix = pixelAddress(src, srcRect, renderWindow.x1, y, srcRowBytes);
        for(int x = renderWindow.x1; x < renderWindow.x2; x++)
        {
            if(srcPix)
            {
                dstPix->r = 255 - srcPix->r;
                dstPix->g = 255 - srcPix->g;
                dstPix->b = 255 - srcPix->b;
                dstPix->a = srcPix->a;
            }
            else
            {
                dstPix->r = 0;
                dstPix->g = 0;
                dstPix->b = 0;
                dstPix->a = 0;
            }

            ++dstPix;
            ++srcPix;
        }
    }

    // we are finished with the source images so release them
    OFX_ASSERT_OK( g_EffectSuite->clipReleaseImage(sourceImg) );
    OFX_ASSERT_OK( g_EffectSuite->clipReleaseImage(outputImg) );
  
    // if we aborted, then we have failed to produce and image, so say so
    if(g_EffectSuite->abort(effectInstance))
        return kOfxStatFailed;

    // otherwise all was well
    return kOfxStatOK;
}

static void SetHost(OfxHost *hostStruct)
{
    g_Host = hostStruct;
}

static OfxStatus OnCreateInstance(OfxImageEffectHandle effectInstance)
{
    OfxPropertySetHandle effectInstanceProperties = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->getPropertySet(effectInstance, &effectInstanceProperties) );

    EffectInstance* instanceObj = new EffectInstance;
    OFX_ASSERT_OK( g_PropertySuite->propSetPointer(effectInstanceProperties, kOfxPropInstanceData, 0, instanceObj) );

    OfxParamSetHandle paramSet;
    OFX_ASSERT_OK( g_EffectSuite->getParamSet(effectInstance, &paramSet) );

    OFX_ASSERT_OK( g_ParameterSuite->paramGetHandle(paramSet, "Size", &instanceObj->SizeParam, 0) );
    OFX_ASSERT_OK( g_ParameterSuite->paramGetHandle(paramSet, "Speed", &instanceObj->SpeedParam, 0) );
    OFX_ASSERT_OK( g_ParameterSuite->paramGetHandle(paramSet, "Intensity", &instanceObj->IntensityParam, 0) );

    return kOfxStatOK;
}

static OfxStatus OnDestroyInstance(OfxImageEffectHandle effectInstance)
{
    OfxPropertySetHandle effectInstanceProperties = nullptr;
    OFX_ASSERT_OK( g_EffectSuite->getPropertySet(effectInstance, &effectInstanceProperties) );

    EffectInstance* instanceObj = nullptr;
    OFX_ASSERT_OK( g_PropertySuite->propGetPointer(effectInstanceProperties, kOfxPropInstanceData, 0, (void**)&instanceObj) );

    delete instanceObj;

    return kOfxStatOK;
}

static OfxStatus OnBeginInstanceChanged(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnEndInstanceChanged(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnInstanceChanged(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnPurgeCaches(OfxImageEffectHandle effectInstance)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnSyncPrivateData(OfxImageEffectHandle effectInstance)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnBeginInstanceEdit(OfxImageEffectHandle effectInstance)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnEndInstanceEdit(OfxImageEffectHandle effectInstance)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnBeginSequenceRender(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnEndSequenceRender(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs)
{
    return kOfxStatReplyDefault; // kOfxStatOK;
}

static OfxStatus OnGetRegionOfDefinition(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus OnGetRegionsOfInterest(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus OnGetFramesNeeded(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus OnIsIdentity(OfxImageEffectHandle effectInstance, OfxPropertySetHandle inArgs, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus OnGetClipPreferences(OfxImageEffectHandle effectInstance, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus OnGetTimeDomain(OfxImageEffectHandle effectInstance, OfxPropertySetHandle outArgs)
{
    return kOfxStatReplyDefault;
}

static OfxStatus MainEntry(const char *action, const void *handle, OfxPropertySetHandle inArgs, OfxPropertySetHandle outArgs)
{
    OfxImageEffectHandle effect = (OfxImageEffectHandle)handle;

    // Called very often.
    if(strcmp(action, kOfxImageEffectActionRender) == 0)
        return OnRender(effect, inArgs);
    else if(strcmp(action, kOfxImageEffectActionBeginSequenceRender) == 0)
        return OnBeginSequenceRender((OfxImageEffectHandle)handle, inArgs);
    else if(strcmp(action, kOfxImageEffectActionEndSequenceRender) == 0)
        return OnEndSequenceRender((OfxImageEffectHandle)handle, inArgs);
    else if(strcmp(action, kOfxImageEffectActionIsIdentity) == 0)
        return OnIsIdentity((OfxImageEffectHandle)handle, inArgs, outArgs);
    // Called often.
    else if(strcmp(action, kOfxActionBeginInstanceChanged) == 0)
        return OnBeginInstanceChanged((OfxImageEffectHandle)handle, inArgs);
    else if(strcmp(action, kOfxActionEndInstanceChanged) == 0)
        return OnEndInstanceChanged((OfxImageEffectHandle)handle, inArgs);
    else if(strcmp(action, kOfxActionInstanceChanged) == 0)
        return OnInstanceChanged((OfxImageEffectHandle)handle, inArgs);
    else if(strcmp(action, kOfxActionPurgeCaches) == 0)
        return OnPurgeCaches((OfxImageEffectHandle)handle);
    else if(strcmp(action, kOfxActionSyncPrivateData) == 0)
        return OnSyncPrivateData((OfxImageEffectHandle)handle);
    else if(strcmp(action, kOfxActionBeginInstanceEdit) == 0)
        return OnBeginInstanceEdit((OfxImageEffectHandle)handle);
    else if(strcmp(action, kOfxActionEndInstanceEdit) == 0)
        return OnEndInstanceEdit((OfxImageEffectHandle)handle);
    else if(strcmp(action, kOfxImageEffectActionGetRegionOfDefinition) == 0)
        return OnGetRegionOfDefinition((OfxImageEffectHandle)handle, inArgs, outArgs);
    else if(strcmp(action, kOfxImageEffectActionGetRegionsOfInterest) == 0)
        return OnGetRegionsOfInterest((OfxImageEffectHandle)handle, inArgs, outArgs);
    else if(strcmp(action, kOfxImageEffectActionGetFramesNeeded) == 0)
        return OnGetFramesNeeded((OfxImageEffectHandle)handle, inArgs, outArgs);
    else if(strcmp(action, kOfxImageEffectActionGetClipPreferences) == 0)
        return OnGetClipPreferences((OfxImageEffectHandle)handle, outArgs);
    else if(strcmp(action, kOfxImageEffectActionGetTimeDomain) == 0)
        return OnGetTimeDomain((OfxImageEffectHandle)handle, outArgs);
    // Called once for entire plugin.
    else if(strcmp(action, kOfxActionLoad) == 0)
        return OnLoad();
    else if(strcmp(action, kOfxActionUnload) == 0)
        return OnUnload();
    else if(strcmp(action, kOfxActionDescribe) == 0)
        return OnDescribe(effect);
    else if(strcmp(action, kOfxImageEffectActionDescribeInContext) == 0)
        return OnDescribeInContext(effect, inArgs);
    else if(strcmp(action, kOfxActionCreateInstance) == 0)
        return OnCreateInstance((OfxImageEffectHandle)handle);
    else if(strcmp(action, kOfxActionDestroyInstance) == 0)
        return OnDestroyInstance((OfxImageEffectHandle)handle);

    return kOfxStatReplyDefault;
}

static OfxPlugin pluginStruct = 
{       
    kOfxImageEffectPluginApi, // pluginApi
    1, // apiVersion
    "info.asawicki.OFX.Test1", // pluginIdentifier
    1, // pluginVersionMajor
    0, // pluginVersionMinor
    SetHost, // setHost
    MainEntry // mainEntry
};

OfxExport int OfxGetNumberOfPlugins()
{       
    return 1;
}

OfxExport OfxPlugin * OfxGetPlugin(int nth)
{
    if(nth == 0)
        return &pluginStruct;
    return 0;
}
